# coding=utf-8
# Copyright 2019 The Routing Transformers Iclr20 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sparse Transformer for benchmarking various sparse attention patterns."""
from __future__ import absolute_import
from __future__ import division

from __future__ import print_function

import numpy as np
from tensor2tensor.layers import common_hparams
from tensor2tensor.utils import registry
from tensor2tensor.utils import t2t_model
import tensorflow as tf

from routing_transformers_iclr20 import utils


@registry.register_model
class SparseTransformer(t2t_model.T2TModel):
  """Sparse Transformer model."""

  @property
  def vocab_size(self):
    return self.problem_hparams.vocab_size["targets"]

  @property
  def batch_size(self):
    return self.hparams.batch_size

  @property
  def hidden_size(self):
    return self.hparams.hidden_size

  @property
  def num_decoder_layers(self):
    return self.hparams.num_decoder_layers

  @property
  def is_decode(self):
    return self.hparams.mode == tf.estimator.ModeKeys.PREDICT

  def process_partial_targets_decoding(self, targets):
    """Processes partially generated targets in decoding mode."""
    targets_shape = utils.shape_list(targets)
    seq_length = targets_shape[1]
    blocks_per_dim = [
        s // q for s, q in zip([seq_length], self.hparams.query_shape)
    ]
    targets = tf.reshape(
        targets, [targets_shape[0], -1,
                  np.prod(self.hparams.query_shape), 1])

    targets = utils.unflatten_blocks_nd(targets, blocks_per_dim)
    targets = utils.put_back_blocks_nd(targets, self.hparams.query_shape)
    targets = tf.reshape(targets, [-1, seq_length])
    return targets

  def prepare_decoder(self, targets):
    """Prepares targets for transformer decoder."""
    shape = utils.shape_list(targets)
    # sequence should be [batch, seq_length]
    assert len(shape) == 2, "Sequence tensors should be 2-dimensional"
    assert len(
        self.hparams.query_shape) == 1, "query shape should be 1-dimensional"

    # Add channel embeddings
    targets = tf.expand_dims(targets, axis=-1)
    targets = utils.get_embeddings(
        targets=targets,
        hidden_size=self.hidden_size,
        vocab_size=self.vocab_size)

    # Shift positions
    targets = utils.right_shift_blockwise_nd(targets, self.hparams.query_shape)
    if self.hparams.add_timing_signal:
      targets += utils.get_timing_signal_1d(self.hparams.max_length,
                                            self.hidden_size)
    return targets

  def multinomial_squeeze(self, logits, temperature=1.0):
    """multinomial sampling from logits."""
    logits_shape = utils.shape_list(logits)
    reshaped_logits = (tf.reshape(logits, [-1, logits_shape[-1]]) / temperature)
    choices = tf.multinomial(reshaped_logits, 1)
    choices = tf.reshape(choices, logits_shape[:-1])
    return tf.to_int32(choices)

  def produce_output(self, decoder_output):
    """Maps decoder output to final logits."""
    # map decoder output to output vocab size
    output = tf.layers.dense(
        decoder_output, self.vocab_size, activation=None, name="final_dense")

    if self.is_decode:
      return output

    return output

  def body(self, features, decode_step=None, cache=None, decoding_stats=None):
    targets = tf.to_int32(features["targets"])
    shape = utils.shape_list(targets)
    # remove the last two dimensions that are always 1.
    targets = tf.reshape(targets, shape[:2])
    # Pad to max_length if needed
    if isinstance(shape[1], tf.Tensor) or shape[1] < self.hparams.max_length:
      targets = tf.pad(targets, [(0, 0),
                                 (0, self.hparams.max_length - shape[1])])
      targets.set_shape([self.batch_size, self.hparams.max_length])
    if self.is_decode:
      targets = self.process_partial_targets_decoding(targets)
    decoder_input = self.prepare_decoder(targets)
    extra_losses = []

    decoder_output = utils.transformer_decoder_layers(
        inputs=decoder_input,
        num_layers=self.num_decoder_layers,
        hparams=self.hparams,
        decode_step=decode_step,
        losses=extra_losses,
        cache=cache,
        name="decoder",
        decoding_stats=decoding_stats)
    logits = self.produce_output(decoder_output)

    # Return logits as-is in decoding mode
    if self.is_decode:
      return logits

    # Add cross entropy loss
    one_hot_targets = tf.one_hot(
        tf.cast(targets, dtype=tf.int32), self.vocab_size)
    x_entropy = tf.nn.softmax_cross_entropy_with_logits_v2(
        labels=one_hot_targets, logits=logits)
    weights = tf.to_float(tf.not_equal(targets, 0))
    loss = tf.reduce_sum(x_entropy * weights) / tf.reduce_sum(weights)

    loss_dict = {"training": loss}
    if extra_losses:
      loss_dict["extra_loss"] = tf.add_n(extra_losses)
    # hack for T2T metrics
    logits = tf.reshape(
        logits,
        utils.shape_list(logits)[:2] + [1, 1] + utils.shape_list(logits)[-1:])
    return logits, loss_dict

  def sample(self, features, decode_step, cache, decoding_stats):
    """Sample step for infer."""
    with tf.variable_scope("sparse_transformer/body", reuse=tf.AUTO_REUSE):
      logits = self.body(features, decode_step, cache, decoding_stats)
      logits = tf.reshape(logits, [self.batch_size, self.vocab_size])
      sample = self.multinomial_squeeze(logits, self.hparams.sampling_temp)
      sample = tf.reshape(sample, [self.batch_size])
      return sample, logits

  def infer(self, features, **kwargs):
    decode_length = self.hparams.max_length
    cache = {}
    decoding_stats = {}
    targets_old = features.get("targets")
    inputs = features.get("inputs")
    start_step = 0
    initial_output = tf.zeros((self.batch_size, decode_length, 1, 1),
                              dtype=tf.int32)

    initial_logits = tf.zeros((self.batch_size, decode_length, self.vocab_size))
    if inputs is not None:
      # decode pipeline primed with a partial input sequence
      start_step = utils.shape_list(inputs)[1]
      inputs = tf.pad(inputs,
                      [[0, 0], [0, decode_length - start_step], [0, 0], [0, 0]])
      inputs.set_shape([self.batch_size, decode_length, 1, 1])
      initial_output = inputs

    # call body once to initialize cache with representations of input frames.
    features["targets"] = initial_output
    with tf.variable_scope("sparse_transformer/body", reuse=tf.AUTO_REUSE):
      self.body(
          features,
          decode_step=None,
          cache=cache,
          decoding_stats=decoding_stats)

    def infer_step(i, recent_output, recent_logits, cache, decoding_stats):
      """Inference step."""
      features_copy = features.copy()
      features_copy["targets"] = recent_output
      cur_sample, cur_logit = self.sample(
          features_copy,
          decode_step=i,
          cache=cache,
          decoding_stats=decoding_stats)
      pos = i
      samples = recent_output + tf.scatter_nd(
          indices=[[b, pos, 0, 0] for b in range(self.batch_size)],
          updates=cur_sample,
          shape=utils.shape_list(recent_output))
      logits = recent_logits + tf.scatter_nd(
          indices=[[b, pos] for b in range(self.batch_size)],
          updates=cur_logit,
          shape=utils.shape_list(recent_logits))
      return i + 1, samples, logits, cache, decoding_stats

    def while_exit_cond(i, result, logits, cache, decoding_stats):
      """Exit the loop if it reaches decode_length."""
      not_overflow = i < decode_length
      return not_overflow

    _, final_result, final_logits, _, decoding_stats = tf.while_loop(
        while_exit_cond,
        infer_step,
        [start_step, initial_output, initial_logits, cache, decoding_stats],
        back_prop=False,
        parallel_iterations=1)

    original_shape = [decode_length]

    blocks_per_dim = [
        s // q for s, q in zip(original_shape, self.hparams.query_shape)
    ]
    final_result_shape = utils.shape_list(final_result)
    final_result = tf.reshape(
        final_result,
        [final_result_shape[0], -1,
         np.prod(self.hparams.query_shape), 1])
    final_logits_shape = utils.shape_list(final_logits)
    final_logits = tf.reshape(final_logits, [
        final_logits_shape[0], -1,
        np.prod(self.hparams.query_shape), final_logits_shape[-1]
    ])
    final_result = utils.unflatten_blocks_nd(final_result, blocks_per_dim)
    final_result = utils.put_back_blocks_nd(final_result,
                                            self.hparams.query_shape)
    final_logits = utils.unflatten_blocks_nd(final_logits, blocks_per_dim)
    final_logits = utils.put_back_blocks_nd(final_logits,
                                            self.hparams.query_shape)

    for name, value in decoding_stats.items():
      tf.summary.scalar("decodes/%s" % name, value / decode_length)

    # Reassign targets back to the previous value.
    if targets_old is not None:
      features["targets"] = targets_old

    return {
        "outputs": final_result,
        "scores": None,
        "logits": final_logits,
        "losses": None,
    }


def target_bottom(x, model_hparams, vocab_size):
  del model_hparams, vocab_size  # Unused
  return x


@registry.register_hparams
def sparse_transformer_local():
  """Set of hyperparameters for a sparse model using only local."""
  hparams = common_hparams.basic_params1()
  hparams.max_length = 4096
  hparams.batch_size = 4096
  hparams.add_hparam("add_timing_signal", False)
  hparams.add_hparam("local_num_heads", 8)
  hparams.add_hparam("sparsity_cluster_num_heads", 0)
  hparams.add_hparam("sparsity_strided_num_heads", 0)
  hparams.add_hparam("sparsity_cluster_strided_num_heads", 0)
  hparams.add_hparam("ema", True)
  hparams.add_hparam("query_shape", (512,))
  hparams.add_hparam("memory_flange", (512,))
  hparams.add_hparam("sparsity_cluster_z_size", 0)
  hparams.add_hparam("sparsity_cluster_attention_window", 0)
  hparams.add_hparam("num_decoder_layers", 24)
  hparams.add_hparam("attention_key_channels", 0)  # Uses hidden_size
  hparams.add_hparam("attention_value_channels", 0)  # Uses hidden_size
  hparams.add_hparam("attention_dropout", 0.0)
  hparams.add_hparam("ffn_layer", "conv_hidden_relu")
  hparams.add_hparam("filter_size", 2048)  # Used in ffn_layer
  hparams.add_hparam("relu_dropout", 0.0)  # Used in ffn_layer
  hparams.bottom = {
      "targets": target_bottom,
  }

  # Optimizer
  hparams.dropout = 0.0
  hparams.clip_grad_norm = 0.  # i.e. no gradient clipping
  hparams.optimizer_adam_epsilon = 1e-9
  hparams.learning_rate_decay_scheme = "noam"
  hparams.learning_rate = 0.1
  hparams.learning_rate_warmup_steps = 6000
  hparams.initializer_gain = 0.2
  hparams.initializer = "uniform_unit_scaling"
  hparams.weight_decay = 0.0
  hparams.optimizer_adam_beta1 = 0.9
  hparams.optimizer_adam_beta2 = 0.98
  hparams.label_smoothing = 0.0
  hparams.summarize_vars = True
  hparams.hidden_size = 512

  # Memory saving measures
  hparams.add_hparam("recompute_grad", False)
  hparams.add_hparam("cache_padding_bias", False)

  # relative attention
  hparams.max_relative_position = 1024
  hparams.add_hparam("local_relative", True)
  hparams.add_hparam("sparsity_cluster_relative", True)
  hparams.add_hparam("sparsity_cluster_strided_relative", True)
  hparams.add_hparam("sparsity_strided_relative", False)

  return hparams


@registry.register_hparams
def enwik8_local8k():
  hparams = sparse_transformer_local()

  hparams.max_length = 8192
  hparams.batch_size = 8192
  hparams.hidden_size = 1024
  hparams.num_decoder_layers = 22
  hparams.local_num_heads = 8
  hparams.filter_size = 3072
  hparams.query_shape = (256,)
  hparams.memory_flange = (256,)
  hparams.max_relative_position = 513
  hparams.attention_dropout = 0.4
  hparams.relu_dropout = 0.4

  return hparams


@registry.register_hparams
def enwik8_local_cluster8k():
  hparams = enwik8_local8k()

  hparams.sparsity_cluster_z_size = 4
  hparams.local_num_heads = 4
  hparams.sparsity_cluster_num_heads = 4
  hparams.sparsity_cluster_attention_window = 512
  hparams.num_decoder_layers = 12
  hparams.attention_dropout = 0.0
  hparams.relu_dropout = 0.0

  return hparams


@registry.register_hparams
def enwik8_local12k():
  hparams = enwik8_local8k()

  hparams.max_length = 12288
  hparams.batch_size = 12288
  hparams.query_shape = (128,)
  hparams.memory_flange = (128,)
  hparams.attention_dropout = 0.4
  hparams.relu_dropout = 0.4
  hparams.max_relative_position = 257
  hparams.hidden_size = 1024

  return hparams


@registry.register_hparams
def enwik8_local_cluster12k():
  """Enwik8 local cluster with 12k sequence length."""
  hparams = enwik8_local12k()

  hparams.sparsity_cluster_z_size = 4
  hparams.local_num_heads = 4
  hparams.sparsity_cluster_num_heads = 4
  hparams.sparsity_cluster_attention_window = 256
  hparams.num_decoder_layers = 12
  hparams.attention_dropout = 0.0
  hparams.relu_dropout = 0.0

  return hparams


@registry.register_hparams
def wikitext103_local4k():
  """Enwik8 local with 4k sequence length."""
  hparams = sparse_transformer_local()
  hparams.max_length = 4096
  hparams.batch_size = 4096
  hparams.hidden_size = 1024
  hparams.filter_size = 3072
  hparams.local_num_heads = 16
  hparams.num_decoder_layers = 16
  hparams.query_shape = (128,)
  hparams.memory_flange = (128,)
  hparams.attention_dropout = 0.0
  hparams.relu_dropout = 0.0
  hparams.max_relative_position = 257
  return hparams


@registry.register_hparams
def wikitext103_local_cluster4k():
  hparams = wikitext103_local4k()
  hparams.local_num_heads = 8
  hparams.sparsity_cluster_num_heads = 8
  hparams.sparsity_cluster_attention_window = 256
  hparams.max_relative_position = 257
  hparams.sparsity_cluster_z_size = 4
  hparams.num_decoder_layers = 10
  return hparams


@registry.register_hparams
def wikitext103_local_cluster_small():
  """Small model."""
  hparams = wikitext103_local_cluster4k()
  hparams.local_num_heads = 4
  hparams.sparsity_cluster_num_heads = 4
  hparams.num_decoder_layers = 10
  hparams.hidden_size = 512
  hparams.batch_size = 2
  hparams.max_length = 512
  hparams.use_fixed_batch_size = True
  hparams.learning_rate = 1e-2
  return hparams
